import { isFunction } from 'underscore';
import CommandAbstract from './view/CommandAbstract';

module.exports = () => {
  let em;
  let c = {};
  const commands = {};
  const defaultCommands = {};
  const defaults = require('./config/config');
  const active = {};
  const add = function(id, obj) {
    if (isFunction(obj)) {
      obj = { run: obj };
    }
    if (!obj.stop) {
      obj.noStop = 1;
    }
    delete obj.initialize;
    obj.id = id;
    commands[id] = CommandAbstract.extend(obj);
    return this;
  };

  return {
    CommandAbstract,
    name: 'Commands',
    init(config) {
      c = config || {};
      for (const name in defaults) {
        if (!(name in c)) c[name] = defaults[name];
      }
      em = c.em;
      const ppfx = c.pStylePrefix;
      if (ppfx) c.stylePrefix = ppfx + c.stylePrefix;
      // Load commands passed via configuration
      for (const k in c.defaults) {
        const obj = c.defaults[k];
        if (obj.id) this.add(obj.id, obj);
      }
      const ViewCode = require('./view/ExportTemplate');
      defaultCommands['select-comp'] = require('./view/SelectComponent');
      defaultCommands['create-comp'] = require('./view/CreateComponent');
      defaultCommands['delete-comp'] = require('./view/DeleteComponent');
      defaultCommands['image-comp'] = require('./view/ImageComponent');
      defaultCommands['move-comp'] = require('./view/MoveComponent');
      defaultCommands['text-comp'] = require('./view/TextComponent');
      defaultCommands['insert-custom'] = require('./view/InsertCustom');
      defaultCommands['export-template'] = ViewCode;
      defaultCommands['sw-visibility'] = require('./view/SwitchVisibility');
      defaultCommands['open-layers'] = require('./view/OpenLayers');
      defaultCommands['open-sm'] = require('./view/OpenStyleManager');
      defaultCommands['open-tm'] = require('./view/OpenTraitManager');
      defaultCommands['open-blocks'] = require('./view/OpenBlocks');
      defaultCommands['open-assets'] = require('./view/OpenAssets');
      defaultCommands['show-offset'] = require('./view/ShowOffset');
      defaultCommands['select-parent'] = require('./view/SelectParent');
      defaultCommands.fullscreen = require('./view/Fullscreen');
      defaultCommands.preview = require('./view/Preview');
      defaultCommands.resize = require('./view/Resize');
      defaultCommands.drag = require('./view/Drag');
      defaultCommands['tlb-delete'] = {
        run(ed) {
          return ed.runCommand('core:component-delete');
        }
      };
      defaultCommands['tlb-clone'] = {
        run(ed) {
          ed.runCommand('core:copy');
          ed.runCommand('core:paste');
        }
      };
      defaultCommands['tlb-move'] = {
        run(ed, sender, opts) {
          let dragger;
          const em = ed.getModel();
          const event = opts && opts.event;
          const sel = ed.getSelected();
          const selAll = [...ed.getSelectedAll()];
          const toolbarStyle = ed.Canvas.getToolbarEl().style;
          const nativeDrag = event && event.type == 'dragstart';
          const defComOptions = { preserveSelected: 1 };
          const hideTlb = () => {
            toolbarStyle.display = 'none';
            em.stopDefault(defComOptions);
          };
          if (!sel || !sel.get('draggable')) {
            console.warn('The element is not draggable');
            return;
          }
          // Without setTimeout the ghost image disappears
          nativeDrag ? setTimeout(() => hideTlb, 0) : hideTlb();

          const onStart = (e, opts) => {
            console.log('start mouse pos ', opts.start);
            console.log('el rect ', opts.elRect);
            const el = opts.el;
            el.style.position = 'absolute';
            el.style.margin = 0;
          };
          const onEnd = e => {
            em.runDefault(defComOptions);
            selAll.forEach(sel => sel.set('status', 'selected'));
            ed.select(selAll);
            sel.emitUpdate();
            if (dragger && dragger.blur) {
              dragger.blur();
            }
          };
          const onDrag = (e, opts) => {
            console.log('Delta ', opts.delta);
            console.log('Current ', opts.current);
          };
          if (em.get('designerMode')) {
            // TODO move grabbing func in editor/canvas from the Sortercommands
            dragger = editor.runCommand('drag', {
              el: sel.view.el,
              options: {
                event,
                onStart,
                onDrag,
                onEnd
              }
            });
          } else {
            if (nativeDrag) {
              event.dataTransfer.setDragImage(sel.view.el, 0, 0);
            }
            const cmdMove = ed.Commands.get('move-comp');
            cmdMove.onEndMoveFromModel = onEnd;
            cmdMove.initSorterFromModels(selAll);
          }
          selAll.forEach(sel => sel.set('status', 'freezed-selected'));
        }
      };
      defaultCommands['core:undo'] = e => e.UndoManager.undo();
      defaultCommands['core:redo'] = e => e.UndoManager.redo();
      [
        ['copy', 'CopyComponent'],
        ['paste', 'PasteComponent'],
        ['component-next', 'ComponentNext'],
        ['component-prev', 'ComponentPrev'],
        ['component-enter', 'ComponentEnter'],
        ['component-exit', 'ComponentExit'],
        ['canvas-clear', 'CanvasClear'],
        ['component-delete', 'ComponentDelete'],
        ['component-style-clear', 'ComponentStyleClear']
      ].forEach(
        item =>
          (defaultCommands[`core:${item[0]}`] = require(`./view/${
            item[1]
          }`).run)
      );
      if (c.em) c.model = c.em.get('Canvas');
      this.loadDefaultCommands();
      return this;
    },
    add,
    get(id) {
      let el = commands[id];

      if (typeof el === 'function') {
        el = new el(c);
        commands[id] = el;
      }

      return el;
    },
    has(id) {
      return !!commands[id];
    },
    getAll() {
      return commands;
    },
    run(id, options = {}) {
      return this.runCommand(this.get(id), options);
    },
    stop(id, options = {}) {
      return this.stopCommand(this.get(id), options);
    },
    isActive(id) {
      return this.getActive().hasOwnProperty(id);
    },
    getActive() {
      return active;
    },
    loadDefaultCommands() {
      for (const id in defaultCommands) {
        this.add(id, defaultCommands[id]);
      }
      return this;
    },
    runCommand(command, options = {}) {
      let result;

      if (command && command.run) {
        const id = command.id;
        const editor = em.get('Editor');
        result = command.callRun(editor, options);

        if (id && command.stop && !command.noStop) {
          active[id] = result;
        }
      }

      return result;
    },
    stopCommand(command, options = {}) {
      let result;
      if (command && command.run) {
        const id = command.id;
        const editor = em.get('Editor');
        result = command.callStop(editor, options);
        if (id) delete active[id];
      }
      return result;
    },
    create(command) {
      const cmd = CommandAbstract.extend(command);
      return new cmd(c);
    }
  };
};
