import { bindAll, isElement, isUndefined } from 'underscore';
import { on, off, getUnitFromValue } from 'utils/mixins';

const ToolbarView = require('dom_components/view/ToolbarView');
const Toolbar = require('dom_components/model/Toolbar');
const key = require('keymaster');
const $ = require('backbone').$;

let showOffsets;

module.exports = {
  init(o) {
    bindAll(this, 'onHover', 'onOut', 'onClick', 'onFrameScroll');
  },
  enable() {
    this.frameOff = this.canvasOff = this.adjScroll = null;
    this.startSelectComponent();
    const em = this.config.em;
    showOffsets = 1;
    em.on('component:update', this.updateAttached, this);
    em.on('change:canvasOffset', this.updateAttached, this);
  },
  startSelectComponent() {
    this.toggleSelectComponent(1);
  },
  stopSelectComponent() {
    this.toggleSelectComponent();
  },
  toggleSelectComponent(enable) {
    const em = this.em;
    const method = enable ? 'on' : 'off';
    const methods = { on, off };
    const body = this.getCanvasBody();
    const win = this.getContentWindow();
    methods[method](body, 'mouseover', this.onHover);
    methods[method](body, 'mouseout', this.onOut);
    methods[method](body, 'click', this.onClick);
    methods[method](win, 'scroll resize', this.onFrameScroll);
    em[method]('component:toggled', this.onSelect, this);
    em[method]('change:componentHovered', this.onHovered, this);
  },
  onHover(e) {
    e.stopPropagation();
    const trg = e.target;
    const $el = $(trg);
    let model = $el.data('model');
    if (!model) {
      let parent = $el.parent();
      while (!model && parent) {
        model = parent.data('model');
        parent = parent.parent();
      }
    }
    if (!this.adjScroll) {
      this.adjScroll = 1;
      this.updateAttached();
    }
    if (model && !model.get('hoverable')) {
      let parent = model && model.parent();
      while (parent && !parent.get('hoverable')) parent = parent.parent();
      model = parent;
    }
    this.em.setHovered(model, { forceChange: 1 });
  },
  onHovered(em, component) {
    const trg = component && component.getEl();
    if (trg) {
      const pos = this.getElementPos(trg);
      this.updateBadge(trg, pos);
      this.updateHighlighter(trg, pos);
      this.showElementOffset(trg, pos);
    }
  },
  onOut(e) {
    e.stopPropagation();
    this.hideBadge();
    this.hideHighlighter();
    this.hideElementOffset();
  },
  showElementOffset(el, pos) {
    const $el = $(el);
    const model = $el.data('model');
    if ((model && model.get('status') == 'selected') || !showOffsets) {
      return;
    }
    this.editor.runCommand('show-offset', {
      el,
      elPos: pos
    });
  },
  hideElementOffset() {
    this.editor.stopCommand('show-offset');
  },
  showFixedElementOffset(el, pos) {
    this.editor.runCommand('show-offset', {
      el,
      elPos: pos,
      state: 'Fixed'
    });
  },
  hideFixedElementOffset() {
    if (this.editor) this.editor.stopCommand('show-offset', { state: 'Fixed' });
  },
  hideHighlighter() {
    this.canvas.getHighlighter().style.display = 'none';
  },
  onClick(e) {
    e.stopPropagation();
    const $el = $(e.target);
    let model = $el.data('model');
    if (!model) {
      let parent = $el.parent();
      while (!model && parent) {
        model = parent.data('model');
        parent = parent.parent();
      }
    }
    if (model) {
      if (model.get('selectable')) {
        this.select(model, e);
      } else {
        let parent = model.parent();
        while (parent && !parent.get('selectable')) {
          parent = parent.parent();
        }
        this.select(parent, e);
      }
    }
  },
  select(model, event = {}) {
    if (!model) return;
    const ctrlKey = event.ctrlKey || event.metaKey;
    const shiftKey = event.shiftKey;
    const { editor } = this;
    const multiple = editor.getConfig('multipleSelection');
    const em = this.em;
    if (ctrlKey && multiple) {
      editor.selectToggle(model);
    } else if (shiftKey && multiple) {
      em.clearSelection(editor.Canvas.getWindow());
      const coll = model.collection;
      const index = coll.indexOf(model);
      let min;
      let max;
      editor.getSelectedAll().forEach(sel => {
        const selColl = sel.collection;
        const selIndex = selColl.indexOf(sel);
        if (selColl === coll) {
          if (selIndex < index) {
            min = isUndefined(min) ? selIndex : Math.max(min, selIndex);
          } else if (selIndex > index) {
            max = isUndefined(max) ? selIndex : Math.min(max, selIndex);
          }
        }
      });
      if (!isUndefined(min)) {
        while (min !== index) {
          editor.selectAdd(coll.at(min));
          min + 1;
        }
      }
      if (!isUndefined(max)) {
        while (max !== index) {
          editor.selectAdd(coll.at(max));
          max - 1;
        }
      }
      editor.selectAdd(model);
    } else {
      editor.select(model);
    }
    this.initResize(model);
  },
  updateBadge(el, pos) {
    const $el = $(el);
    const canvas = this.canvas;
    const config = canvas.getConfig();
    const customeLabel = config.customBadgeLabel;
    this.cacheEl = el;
    const model = $el.data('model');
    if (!model || !model.get('badgable')) {
      return;
    }
    const badge = this.getBadge();
    let badgeLabel = model.getIcon() + model.getName();
    badgeLabel = customeLabel ? customeLabel(model) : badgeLabel;
    badge.innerHTML = badgeLabel;
    const bStyle = badge.style;
    const u = 'px';
    bStyle.display = 'block';
    const canvasPos = canvas.getCanvasView().getPosition();
    if (canvasPos) {
      const badgeH = badge ? badge.offsetHeight : 0;
      const badgeW = badge ? badge.offsetWidth : 0;
      const top =
        pos.top - badgeH < canvasPos.top ? canvasPos.top : pos.top - badgeH;
      const left =
        pos.left + badgeW < canvasPos.left ? canvasPos.left : pos.left;
      bStyle.top = top + u;
      bStyle.left = left + u;
    }
  },
  updateHighlighter(el, pos) {
    const $el = $(el);
    const model = $el.data('model');
    if (
      !model ||
      !model.get('hoverable') ||
      model.get('status') == 'selected'
    ) {
      return;
    }
    const hlEl = this.canvas.getHighlighter();
    const hlStyle = hlEl.style;
    const unit = 'px';
    hlStyle.left = pos.left + unit;
    hlStyle.top = pos.top + unit;
    hlStyle.height = pos.height + unit;
    hlStyle.width = pos.width + unit;
    hlStyle.display = 'block';
  },
  onSelect() {
    const model = this.em.getSelected();
    const view = model && model.view;
    this.updateToolbar(model);
    if (view) {
      const { el } = view;
      this.showFixedElementOffset(el);
      this.hideElementOffset();
      this.hideHighlighter();
      this.initResize(el);
    } else {
      this.editor.stopCommand('resize');
    }
  },
  initResize(elem) {
    const em = this.em;
    const editor = em ? em.get('Editor') : '';
    const config = em ? em.get('Config') : '';
    const pfx = config.stylePrefix || '';
    const resizeClass = `${pfx}resizing`;
    const model = !isElement(elem) ? elem : em.getSelected();
    const resizable = model.get('resizable');
    const el = isElement(elem) ? elem : model.getEl();
    let options = {};
    let modelToStyle;
    const toggleBodyClass = (method, e, opts) => {
      const docs = opts.docs;
      if (docs) {
        docs.forEach(doc => {
          const body = doc.body;
          const cls = body.className || '';
          body.className = (method == 'add'
            ? `${cls} ${resizeClass}`
            : cls.replace(resizeClass, '')
          ).trim();
        });
      }
    };
    if (editor && resizable) {
      options = {
        onStart(e, opts = {}) {
          const { el, config, resizer } = opts;
          const {
            keyHeight,
            keyWidth,
            currentUnit,
            keepAutoHeight,
            keepAutoWidth
          } = config;
          toggleBodyClass('add', e, opts);
          modelToStyle = em.get('StyleManager').getModelToStyle(model);
          const computedStyle = getComputedStyle(el);
          const modelStyle = modelToStyle.getStyle();
          let currentWidth = modelStyle[keyWidth];
          config.autoWidth = keepAutoWidth && currentWidth === 'auto';
          if (isNaN(parseFloat(currentWidth))) {
            currentWidth = computedStyle[keyWidth];
          }
          let currentHeight = modelStyle[keyHeight];
          config.autoHeight = keepAutoHeight && currentHeight === 'auto';
          if (isNaN(parseFloat(currentHeight))) {
            currentHeight = computedStyle[keyHeight];
          }
          resizer.startDim.w = parseFloat(currentWidth);
          resizer.startDim.h = parseFloat(currentHeight);
          showOffsets = 0;
          if (currentUnit) {
            config.unitHeight = getUnitFromValue(currentHeight);
            config.unitWidth = getUnitFromValue(currentWidth);
          }
        },
        onMove() {
          editor.trigger('change:canvasOffset');
        },
        onEnd(e, opts) {
          toggleBodyClass('remove', e, opts);
          editor.trigger('change:canvasOffset');
          showOffsets = 1;
        },
        updateTarget(el, rect, options = {}) {
          if (!modelToStyle) {
            return;
          }
          const { store, selectedHandler, config } = options;
          const {
            keyHeight,
            keyWidth,
            autoHeight,
            autoWidth,
            unitWidth,
            unitHeight
          } = config;
          const onlyHeight = ['tc', 'bc'].indexOf(selectedHandler) >= 0;
          const onlyWidth = ['cl', 'cr'].indexOf(selectedHandler) >= 0;
          const style = modelToStyle.getStyle();
          if (!onlyHeight) {
            style[keyWidth] = autoWidth ? 'auto' : `${rect.w}${unitWidth}`;
          }
          if (!onlyWidth) {
            style[keyHeight] = autoHeight ? 'auto' : `${rect.h}${unitHeight}`;
          }
          modelToStyle.setStyle(style, { avoidStore: 1 });
          const updateEvent = `update:component:style`;
          if (em && em.trigger) {
            em.trigger(
              `${updateEvent}:${keyHeight} ${updateEvent}:${keyWidth}`
            );
          }
          if (store) {
            modelToStyle.trigger('change:style', modelToStyle, style, {});
          }
        }
      };
      if (typeof resizable === 'object') {
        options = { ...options, ...resizable };
      }
      editor.runCommand('resize', { el, options });
    } else {
      editor.stopCommand('resize');
    }
  },
  updateToolbar(mod) {
    const em = this.config.em;
    const model = mod == em ? em.getSelected() : mod;
    const toolbarEl = this.canvas.getToolbarEl();
    const toolbarStyle = toolbarEl.style;
    if (!model) {
      toolbarStyle.opacity = 0;
      return;
    }
    const toolbar = model.get('toolbar');
    const ppfx = this.ppfx;
    const showToolbar = em.get('Config').showToolbar;
    if (showToolbar && toolbar && toolbar.length) {
      toolbarStyle.opacity = '';
      toolbarStyle.display = '';
      if (!this.toolbar) {
        toolbarEl.innerHTML = '';
        this.toolbar = new Toolbar(toolbar);
        const toolbarView = new ToolbarView({
          collection: this.toolbar,
          editor: this.editor
        });
        toolbarEl.appendChild(toolbarView.render().el);
      }
      this.toolbar.reset(toolbar);
      const view = model.view;
      toolbarStyle.top = '-100px';
      toolbarStyle.left = 0;
      setTimeout(() => view && this.updateToolbarPos(view.el), 0);
    } else {
      toolbarStyle.display = 'none';
    }
  },
  updateToolbarPos(el, elPos) {
    const unit = 'px';
    const toolbarEl = this.canvas.getToolbarEl();
    const toolbarStyle = toolbarEl.style;
    toolbarStyle.opacity = 0;
    const pos = this.canvas.getTargetToElementDim(toolbarEl, el, {
      elPos,
      event: 'toolbarPosUpdate'
    });
    if (pos) {
      const leftPos = pos.left + pos.elementWidth - pos.targetWidth;
      toolbarStyle.top = pos.top + unit;
      toolbarStyle.left = (leftPos < 0 ? 0 : leftPos) + unit;
      toolbarStyle.opacity = '';
    }
  },
  getCanvasPosition() {
    return this.canvas.getCanvasView().getPosition();
  },
  clean() {
    if (this.selEl) this.selEl.removeClass(this.hoverClass);
  },
  getBadge() {
    return this.canvas.getBadgeEl();
  },
  onFrameScroll() {
    const el = this.cacheEl;
    if (el) {
      const elPos = this.getElementPos(el);
      this.updateBadge(el, elPos);
      const model = this.em.getSelected();
      if (model) {
        this.updateToolbarPos(model.view.el);
      }
    }
  },
  updateAttached() {
    const model = this.em.getSelected();
    const view = model && model.view;
    if (view) {
      const { el } = view;
      this.updateToolbarPos(el);
      this.showFixedElementOffset(el);
    }
  },
  getElementPos(el, badge) {
    return this.canvas.getCanvasView().getElementPos(el);
  },
  hideBadge() {
    this.getBadge().style.display = 'none';
  },
  cleanPrevious(model) {
    if (model && model.set) {
      model.set({
        status: '',
        state: ''
      });
    }
  },
  getContentWindow() {
    return this.frameEl.contentWindow;
  },
  run(editor) {
    this.editor = editor && editor.get('Editor');
    this.enable();
    this.onSelect();
  },
  stop(editor, sender, opts = {}) {
    const em = this.em;
    this.stopSelectComponent();
    if (!opts.preserveSelected) {
      em.setSelected(null);
    }
    this.clean();
    this.hideBadge();
    this.hideFixedElementOffset();
    this.canvas.getToolbarEl().style.display = 'none';
    em.off('component:update', this.updateAttached, this);
    em.off('change:canvasOffset', this.updateAttached, this);
  }
};
