import { isArray } from 'underscore';

module.exports = () => {
  let em;
  let c = {};
  const defaults = require('./config/config');
  const CssRule = require('./model/CssRule');
  const CssRules = require('./model/CssRules');
  const CssRulesView = require('./view/CssRulesView');
  const Selectors = require('selector_manager/model/Selectors');
  const Selector = require('selector_manager/model/Selector');

  let rules;
  let rulesView;

  return {
    Selectors,
    name: 'CssComposer',
    storageKey() {
      const keys = [];
      const smc = (c.stm && c.stm.getConfig()) || {};
      if (smc.storeCss) keys.push('css');
      if (smc.storeStyles) keys.push('styles');
      return keys;
    },
    init(config) {
      c = config || {};
      for (const name in defaults) {
        if (!(name in c)) c[name] = defaults[name];
      }
      const ppfx = c.pStylePrefix;
      if (ppfx) {
        c.stylePrefix = ppfx + c.stylePrefix;
      }
      const elStyle = (c.em && c.em.config.style) || '';
      c.rules = elStyle || c.rules;
      em = c.em;
      rules = new CssRules([], c);
      rulesView = new CssRulesView({
        collection: rules,
        config: c
      });
      return this;
    },
    onLoad() {
      rules.add(c.rules);
    },
    postLoad(em) {
      const ev = 'add remove';
      const rules = this.getAll();
      const um = em.get('UndoManager');
      if (um && um.add) {
        um.add(rules);
      }
      em.stopListening(rules, ev, this.handleChange);
      em.listenTo(rules, ev, this.handleChange);
      rules.each(rule => this.handleChange(rule, { avoidStore: 1 }));
    },
    handleChange(model, opts = {}) {
      const ev = 'change:style';
      const um = em.get('UndoManager');
      if (um && um.add) {
        um.add(model);
      }
      const handleUpdates = em.handleUpdates.bind(em);
      em.stopListening(model, ev, handleUpdates);
      em.listenTo(model, ev, handleUpdates);
      if (!opts.avoidStore) {
        handleUpdates('', '', opts);
      }
    },
    load(data) {
      let d = data || '';
      if (!d && c.stm) {
        d = c.em.getCacheLoad();
      }
      let obj = d.styles || '';
      if (d.styles) {
        try {
          obj = JSON.parse(d.styles);
        } catch (err) {
          console.log('error load css composer');
        }
      } else if (d.css) {
        obj = c.em.get('Parser').parseCss(d.css);
      }
      if (isArray(obj)) {
        if (obj.length) {
          rules.reset(obj);
        }
      } else if (obj) {
        rules.reset(obj);
      }

      return obj;
    },
    store(noStore) {
      if (!c.stm) {
        return;
      }
      const obj = {};
      const keys = this.storageKey();
      if (keys.indexOf('css') >= 0) obj.css = c.em.getCss();
      if (keys.indexOf('styles') >= 0) obj.styles = JSON.stringify(rules);
      if (!noStore) c.stm.store(obj);
      return obj;
    },
    add(selectors, state, width, opts = {}) {
      const s = state || '';
      const w = width || '';
      const opt = { ...opts };
      let rule = this.get(selectors, s, w, opt);
      if (rule && rule.config && !rule.config.singleAtRule) {
        return rule;
      }
      opt.state = s;
      opt.mediaText = w;
      opt.selectors = '';
      rule = new CssRule(opt, c);
      rule.get('selectors').add(selectors);
      rules.add(rule);
      return rule;
    },
    get(selectors, state, width, ruleProps) {
      let rule = null;
      rules.each(m => {
        if (rule) return;
        if (m.compare(selectors, state, width, ruleProps)) rule = m;
      });
      return rule;
    },
    getAll() {
      return rules;
    },
    clear() {
      this.getAll().reset();
      return this;
    },
    addCollection(data, opts = {}) {
      const result = [];
      const d = data instanceof Array ? data : [data];
      for (let i = 0, l = d.length; i < l; i++) {
        const rule = d[i] || {};
        if (!rule.selectors) {
          continue;
        }
        const sm = c.em && c.em.get('SelectorManager');
        if (!sm) console.warn('Selector Manager not found');
        const sl = rule.selectors;
        const sels = sl instanceof Array ? sl : [sl];
        const newSels = [];
        for (let j = 0, le = sels.length; j < le; j++) {
          const selec = sm.add(sels[j]);
          newSels.push(selec);
        }
        const modelExists = this.get(newSels, rule.state, rule.mediaText, rule);
        const model = this.add(newSels, rule.state, rule.mediaText, rule);
        const updateStyle = !modelExists || !opts.avoidUpdateStyle;
        const style = rule.style || {};
        if (updateStyle) {
          const styleUpdate = opts.extend
            ? { ...model.get('style'), ...style }
            : style;
          model.set('style', styleUpdate);
        }
        result.push(model);
      }
      return result;
    },
    setIdRule(name, style = {}, opts = {}) {
      const state = opts.state || '';
      const media = opts.mediaText || em.getCurrentMedia();
      const sm = em.get('SelectorManager');
      const selector = sm.add({ name, type: Selector.TYPE_ID });
      const rule = this.add(selector, state, media);
      rule.setStyle(style, opts);
      return rule;
    },
    getIdRule(name, opts = {}) {
      const state = opts.state || '';
      const media = opts.mediaText || em.getCurrentMedia();
      const selector = em.get('SelectorManager').get(name, Selector.TYPE_ID);
      return selector && this.get(selector, state, media);
    },
    setClassRule(name, style = {}, opts = {}) {
      const state = opts.state || '';
      const media = opts.mediaText || em.getCurrentMedia();
      const sm = em.get('SelectorManager');
      const selector = sm.add({ name, type: Selector.TYPE_CLASS });
      const rule = this.add(selector, state, media);
      rule.setStyle(style, opts);
      return rule;
    },
    getClassRule(name, opts = {}) {
      const state = opts.state || '';
      const media = opts.mediaText || em.getCurrentMedia();
      const selector = em.get('SelectorManager').get(name, Selector.TYPE_CLASS);
      return selector && this.get(selector, state, media);
    },
    render() {
      return rulesView.render().el;
    }
  };
};
