import Backbone from 'backbone';
import { isEmpty, isObject, isArray } from 'underscore';

module.exports = () => {
  let c = {};
  let em;
  const defaults = require('./config/config');
  const Component = require('./model/Component');
  const ComponentView = require('./view/ComponentView');
  const Components = require('./model/Components');
  const ComponentsView = require('./view/ComponentsView');

  let component;
  let componentView;
  const componentTypes = [
    {
      id: 'cell',
      model: require('./model/ComponentTableCell'),
      view: require('./view/ComponentTableCellView')
    },
    {
      id: 'row',
      model: require('./model/ComponentTableRow'),
      view: require('./view/ComponentTableRowView')
    },
    {
      id: 'table',
      model: require('./model/ComponentTable'),
      view: require('./view/ComponentTableView')
    },
    {
      id: 'thead',
      model: require('./model/ComponentTableHead'),
      view: require('./view/ComponentTableHeadView')
    },
    {
      id: 'tbody',
      model: require('./model/ComponentTableBody'),
      view: require('./view/ComponentTableBodyView')
    },
    {
      id: 'tfoot',
      model: require('./model/ComponentTableFoot'),
      view: require('./view/ComponentTableFootView')
    },
    {
      id: 'map',
      model: require('./model/ComponentMap'),
      view: require('./view/ComponentMapView')
    },
    {
      id: 'link',
      model: require('./model/ComponentLink'),
      view: require('./view/ComponentLinkView')
    },
    {
      id: 'label',
      model: require('./model/ComponentLabel'),
      view: require('./view/ComponentLabelView')
    },
    {
      id: 'video',
      model: require('./model/ComponentVideo'),
      view: require('./view/ComponentVideoView')
    },
    {
      id: 'image',
      model: require('./model/ComponentImage'),
      view: require('./view/ComponentImageView')
    },
    {
      id: 'script',
      model: require('./model/ComponentScript'),
      view: require('./view/ComponentScriptView')
    },
    {
      id: 'svg',
      model: require('./model/ComponentSvg'),
      view: require('./view/ComponentSvgView')
    },
    {
      id: 'textnode',
      model: require('./model/ComponentTextNode'),
      view: require('./view/ComponentTextNodeView')
    },
    {
      id: 'text',
      model: require('./model/ComponentText'),
      view: require('./view/ComponentTextView')
    },
    {
      id: 'wrapper',
      model: require('./model/ComponentWrapper'),
      view: ComponentView
    },
    {
      id: 'default',
      model: Component,
      view: ComponentView
    }
  ];

  return {
    Component,
    Components,
    ComponentsView,
    componentTypes,
    name: 'DomComponents',
    getConfig() {
      return c;
    },
    storageKey() {
      const keys = [];
      const smc = (c.stm && c.stm.getConfig()) || {};
      if (smc.storeHtml) keys.push('html');
      if (smc.storeComponents) keys.push('components');
      return keys;
    },
    init(config) {
      c = config || {};
      em = c.em;
      this.em = em;

      if (em) {
        c.components = em.config.components || c.components;
      }
      for (const name in defaults) {
        if (!(name in c)) c[name] = defaults[name];
      }
      const ppfx = c.pStylePrefix;
      if (ppfx) {
        c.stylePrefix = ppfx + c.stylePrefix;
      }
      if (em) {
        c.modal = em.get('Modal') || '';
        c.am = em.get('AssetManager') || '';
        em.get('Parser').compTypes = componentTypes;
        em.on('change:componentHovered', this.componentHovered, this);
        const selected = em.get('selected');
        em.listenTo(selected, 'add', (sel, c, opts) =>
          this.selectAdd(sel, opts)
        );
        em.listenTo(selected, 'remove', (sel, c, opts) =>
          this.selectRemove(sel, opts)
        );
      }
      let components = c.components;
      let wrapper = { ...c.wrapper };
      wrapper['custom-name'] = c.wrapperName;
      wrapper.wrapper = 1;
      wrapper.type = 'wrapper';
      if (
        components &&
        components.constructor === Object &&
        components.wrapper
      ) {
        wrapper = { ...components };
        components = components.components || [];
        wrapper.components = [];
        if (em) {
          em.config.components = components;
          c.components = components;
        }
      }
      component = new Component(wrapper, {
        em,
        config: c,
        componentTypes
      });
      component.set({ attributes: { id: 'wrapper' } });
      componentView = new ComponentView({
        model: component,
        config: c,
        componentTypes
      });
      return this;
    },
    onLoad() {
      this.setComponents(c.components);
    },
    postLoad() {
      this.handleChanges(this.getWrapper(), null, { avoidStore: 1 });
    },
    handleChanges(model, value, opts = {}) {
      const comps = model.components();
      const um = em.get('UndoManager');
      const handleUpdates = em.handleUpdates.bind(em);
      const handleChanges = this.handleChanges.bind(this);
      const handleChangesColl = this.handleChangesColl.bind(this);
      const handleRemoves = this.handleRemoves.bind(this);
      if (um && um.add) {
        um.add(model);
      }
      if (um && comps && em.add) {
        um.add(comps);
      }
      const evn = 'change:style change:content change:attributes change:src';
      [
        [model, evn, handleUpdates],
        [model, 'change:components', handleChangesColl],
        [comps, 'add', handleChanges],
        [comps, 'remove', handleRemoves],
        [model.get('classes'), 'add remove', handleUpdates]
      ].forEach(els => {
        em.stopListening(els[0], els[1], els[2]);
        em.listenTo(els[0], els[1], els[2]);
      });
      if (!opts.avoidStore) {
        handleUpdates('', '', opts);
      }
      comps.each(model => this.handleChanges(model, value, opts));
    },
    handleChangesColl(model, coll) {
      const um = em.get('UndoManager');
      if (um && coll instanceof Backbone.Collection) {
        const handleChanges = this.handleChanges.bind(this);
        const handleRemoves = this.handleRemoves.bind(this);
        um.add(coll);
        [[coll, 'add', handleChanges], [coll, 'remove', handleRemoves]].forEach(
          els => {
            em.stopListening(els[0], els[1], els[2]);
            em.listenTo(els[0], els[1], els[2]);
          }
        );
      }
    },
    handleRemoves(model, value, opts = {}) {
      if (!opts.avoidStore) {
        em.handleUpdates(model, value, opts);
      }
    },
    load(data = '') {
      const { em } = this;
      let result = '';
      if (!data && c.stm) {
        data = c.em.getCacheLoad();
      }
      const { components, html } = data;
      if (components) {
        if (isObject(components) || isArray(components)) {
          result = components;
        } else {
          try {
            result = JSON.parse(components);
          } catch (err) {
            if (em && em.logError) {
              em.logError(err);
            }
          }
        }
      } else if (html) {
        result = html;
      }
      const isObj = result && result.constructor === Object;
      if ((result && result.length) || isObj) {
        this.clear();
        if (isObj) {
          this.getWrapper().set(result);
        } else {
          this.getComponents().add(result);
        }
      }
      return result;
    },
    store(noStore) {
      if (!c.stm) {
        return;
      }
      const obj = {};
      const keys = this.storageKey();
      if (keys.indexOf('html') >= 0) {
        obj.html = c.em.getHtml();
      }
      if (keys.indexOf('components') >= 0) {
        const { em } = this;
        const storeWrap = c.storeWrapper;
        const toStore = storeWrap ? this.getWrapper() : this.getComponents();
        obj.components = JSON.stringify(toStore);
      }
      if (!noStore) {
        c.stm.store(obj);
      }
      return obj;
    },
    getComponent() {
      return component;
    },
    getWrapper() {
      return this.getComponent();
    },
    getComponents() {
      return this.getWrapper().get('components');
    },
    addComponent(component) {
      return this.getComponents().add(component);
    },
    render() {
      return componentView.render().el;
    },
    clear() {
      this.getComponents().reset();
      return this;
    },
    setComponents(components) {
      this.clear().addComponent(components);
    },
    addType(type, methods) {
      const {
        model = {},
        view = {},
        isComponent,
        extend,
        extendView
      } = methods;
      const compType = this.getType(type);
      const extendType = this.getType(extend);
      const extendViewType = this.getType(extendView);
      const typeToExtend = extendType || (compType || this.getType('default'));
      const modelToExt = typeToExtend.model;
      const viewToExt = extendViewType
        ? extendViewType.view
        : typeToExtend.view;
      if (typeof model === 'object') {
        methods.model = modelToExt.extend(
          {
            ...model,
            defaults: {
              ...modelToExt.prototype.defaults,
              ...(model.defaults || {})
            }
          },
          {
            isComponent:
              compType && !extendType && !isComponent
                ? modelToExt.isComponent
                : isComponent || (() => 0)
          }
        );
      }
      if (typeof view === 'object') {
        methods.view = viewToExt.extend({ ...view });
      }
      if (compType) {
        compType.model = methods.model;
        compType.view = methods.view;
      } else {
        methods.id = type;
        componentTypes.unshift(methods);
      }
      return this;
    },
    getType(type) {
      const df = componentTypes;
      for (let it = 0; it < df.length; it++) {
        const dfId = df[it].id;
        if (dfId == type) {
          return df[it];
        }
      }
    },
    getTypes() {
      return componentTypes;
    },
    selectAdd(component, opts = {}) {
      if (component) {
        component.set({
          status: 'selected'
        });
        ['component:selected', 'component:toggled'].forEach(event =>
          this.em.trigger(event, component, opts)
        );
      }
    },
    selectRemove(component, opts = {}) {
      if (component) {
        const { em } = this;
        component.set({
          status: '',
          state: ''
        });
        ['component:deselected', 'component:toggled'].forEach(event =>
          this.em.trigger(event, component, opts)
        );
      }
    },
    componentHovered() {
      const em = c.em;
      const model = em.get('componentHovered');
      const previous = em.previous('componentHovered');
      const state = 'hovered';
      if (previous && previous.get('status') == state) {
        previous.set({
          status: '',
          state: ''
        });
      }
      if (model && isEmpty(model.get('status'))) {
        model.set('status', state);
      }
    }
  };
};
