import {
  isUndefined,
  isFunction,
  isObject,
  isArray,
  isEmpty,
  isBoolean,
  has,
  isString,
  forEach,
  result,
  keys
} from 'underscore';
import { shallowDiff, hasDnd } from 'utils/mixins';
import Styleable from 'domain_abstract/model/Styleable';

const Backbone = require('backbone');
const Components = require('./Components');
const Selector = require('selector_manager/model/Selector');
const Selectors = require('selector_manager/model/Selectors');
const Traits = require('trait_manager/model/Traits');

const componentList = {};
let componentIndex = 0;

const escapeRegExp = str => str.replace(/[|\\{}()[\]^$+*?.]/g, '\\$&');

const avoidInline = em => em && em.getConfig('avoidInlineStyle');

const Component = Backbone.Model.extend(Styleable).extend(
  {
    defaults: {
      tagName: 'div',
      type: '',
      name: '',
      removable: true,
      draggable: true,
      droppable: true,
      badgable: true,
      stylable: true,
      'stylable-require': '',
      'style-signature': '',
      unstylable: '',
      highlightable: true,
      copyable: true,
      resizable: false,
      editable: false,
      layerable: true,
      selectable: true,
      hoverable: true,
      void: false,
      state: '',
      status: '',
      content: '',
      icon: '',
      style: '',
      classes: '',
      script: '',
      attributes: '',
      traits: ['id', 'title'],
      propagate: '',
      toolbar: null,
      parent: null
    },
    init() {},
    updated() {},
    removed() {},
    initialize(props = {}, opt = {}) {
      const em = opt.em;
      const parent = this.parent();
      const parentAttr = parent && parent.attributes;
      if (parentAttr && parentAttr.propagate) {
        let newAttr = {};
        const toPropagate = parentAttr.propagate;
        toPropagate.forEach(prop => (newAttr[prop] = parent.get(prop)));
        newAttr.propagate = toPropagate;
        newAttr = { ...newAttr, ...props };
        this.set(newAttr);
      }
      const propagate = this.get('propagate');
      if (propagate) {
        this.set('propagate', isArray(propagate) ? propagate : [propagate]);
      }
      if (
        opt &&
        opt.config &&
        opt.config.voidElements.indexOf(this.get('tagName')) >= 0
      ) {
        this.set('void', true);
      }
      opt.em = em;
      this.opt = opt;
      this.em = em;
      this.config = opt.config || {};
      this.ccid = Component.createId(this);
      this.set('attributes', {
        ...(this.defaults.attributes || {}),
        ...(this.get('attributes') || {})
      });
      this.initClasses();
      this.initTraits();
      this.initComponents();
      this.initToolbar();
      this.listenTo(this, 'change:script', this.scriptUpdated);
      this.listenTo(this, 'change:tagName', this.tagUpdated);
      this.listenTo(this, 'change:attributes', this.attrUpdated);
      this.set('status', '');
      ['classes', 'traits', 'components'].forEach(name => {
        const events = `add remove ${name !== 'components' ? 'change' : ''}`;
        this.listenTo(this.get(name), events.trim(), (...args) =>
          this.emitUpdate(name, ...args)
        );
      });
      if (!opt.temporary) {
        this.init();
        if (em && em.trigger) {
          em.trigger('component:create', this);
        }
      }
    },
    is(type) {
      return !!(this.get('type') == type);
    },
    index() {
      const { collection } = this;
      return collection && collection.indexOf(this);
    },
    find(query) {
      const result = [];
      this.view.$el.find(query).each((el, i, $els) => {
        const $el = $els.eq(i);
        const model = $el.data('model');
        if (model) {
          result.push(model);
        }
      });
      return result;
    },
    closest(query) {
      const result = this.view.$el.closest(query);
      return result.length && result.data('model');
    },
    tagUpdated() {
      const coll = this.collection;
      const at = coll.indexOf(this);
      coll.remove(this);
      coll.add(this, { at });
    },
    replaceWith(el) {
      const coll = this.collection;
      const at = coll.indexOf(this);
      coll.remove(this);
      return coll.add(el, { at });
    },
    attrUpdated() {
      this.setAttributes(this.get('attributes'), { silent: 1 });
    },
    setAttributes(attrs, opts = {}) {
      attrs = { ...attrs };
      const classes = attrs.class;
      if (classes) {
        this.setClass(classes);
      }
      delete attrs.class;
      const style = attrs.style;
      if (style) {
        this.setStyle(style);
      }
      delete attrs.style;
      this.set('attributes', attrs, opts);
      const attrPrev = { ...this.previous('attributes') };
      const diff = shallowDiff(attrPrev, attrs);
      keys(diff).forEach(pr => this.trigger(`change:attributes:${pr}`));
      return this;
    },
    addAttributes(attrs) {
      const newAttrs = { ...this.getAttributes(), ...attrs };
      this.setAttributes(newAttrs);

      return this;
    },
    getStyle() {
      const em = this.em;
      if (em && em.getConfig('avoidInlineStyle')) {
        const state = this.get('state');
        const cc = em.get('CssComposer');
        const rule = cc.getIdRule(this.getId(), { state });
        this.rule = rule;
        if (rule) {
          return rule.getStyle();
        }
      }
      return Styleable.getStyle.call(this);
    },
    setStyle(prop = {}, opts = {}) {
      const em = this.em;
      const { opt } = this;
      if (em && em.getConfig('avoidInlineStyle') && !opt.temporary) {
        prop = isString(prop) ? this.parseStyle(prop) : prop;
        prop = { ...prop, ...this.get('style') };
        const state = this.get('state');
        const cc = em.get('CssComposer');
        const propOrig = this.getStyle();
        this.rule = cc.setIdRule(this.getId(), prop, { ...opts, state });
        const diff = shallowDiff(propOrig, prop);
        this.set('style', {}, { silent: 1 });
        keys(diff).forEach(pr => this.trigger(`change:style:${pr}`));
      } else {
        prop = Styleable.setStyle.apply(this, arguments);
      }
      return prop;
    },
    getAttributes() {
      const { em } = this;
      const classes = [];
      const attributes = { ...this.get('attributes') };
      const sm = em && em.get('SelectorManager');
      const id = this.getId();
      this.get('classes').forEach(cls =>
        classes.push(isString(cls) ? cls : cls.get('name'))
      );
      if (classes.length) {
        attributes.class = classes.join(' ');
      }
      if (!has(attributes, 'id')) {
        let hasStyle;
        if (avoidInline(em)) {
          hasStyle = sm && sm.get(id, sm.Selector.TYPE_ID);
        } else if (!isEmpty(this.getStyle())) {
          hasStyle = 1;
        }
        if (hasStyle) {
          attributes.id = this.getId();
        }
      }
      return attributes;
    },
    addClass(classes) {
      const added = this.em.get('SelectorManager').addClass(classes);
      return this.get('classes').add(added);
    },
    setClass(classes) {
      this.get('classes').reset();
      return this.addClass(classes);
    },
    removeClass(classes) {
      const removed = [];
      classes = isArray(classes) ? classes : [classes];
      const selectors = this.get('classes');
      const type = Selector.TYPE_CLASS;

      classes.forEach(classe => {
        const classes = classe.split(' ');
        classes.forEach(name => {
          const selector = selectors.where({ name, type })[0];
          if (selector) {
            removed.push(selectors.remove(selector));
          }
        });
      });
      return removed;
    },
    getClasses() {
      const attr = this.getAttributes();
      const classStr = attr.class;
      return classStr ? classStr.split(' ') : [];
    },
    initClasses() {
      const event = 'change:classes';
      const toListen = [this, event, this.initClasses];
      this.stopListening(...toListen);
      const classes = this.normalizeClasses(this.get('classes') || []);
      const selectors = new Selectors([]);
      this.set('classes', selectors);
      selectors.add(classes);
      this.listenTo(...toListen);
      return this;
    },
    initComponents() {
      const event = 'change:components';
      const toListen = [this, event, this.initComponents];
      this.stopListening(...toListen);
      const comps = new Components(null, this.opt);
      comps.parent = this;
      const components = this.get('components');
      const addChild = !this.opt.avoidChildren;
      this.set('components', comps);
      if (addChild) {
        comps.add(components);
      }
      this.listenTo(...toListen);
      return this;
    },
    initTraits() {
      const event = 'change:traits';
      const toListen = [this, event, this.initTraits];
      this.stopListening(...toListen);
      this.loadTraits();
      const attrs = { ...this.get('attributes') };
      const traits = this.get('traits');
      traits.each(trait => {
        if (!trait.get('changeProp')) {
          const name = trait.get('name');
          const value = trait.getInitValue();
          if (name && value) attrs[name] = value;
        }
      });
      if (traits.length) {
        this.set('attributes', attrs);
      }
      this.listenTo(...toListen);
      return this;
    },
    append(components, opts = {}) {
      const result = this.components().add(components, opts);
      return isArray(result) ? result : [result];
    },
    components(components) {
      const coll = this.get('components');
      if (isUndefined(components)) {
        return coll;
      }
      coll.reset();
      return components && this.append(components);
    },
    parent() {
      const coll = this.collection;
      return coll && coll.parent;
    },
    scriptUpdated() {
      this.set('scriptUpdated', 1);
    },
    validResizeParent(classParent = null) {
      let r = false;
      if (this.parent() && this.parent().getClasses() && classParent) {
        r = this.parent()
          .getClasses()
          .find(element => {
            const resizeClass = new RegExp(`${classParent}`, 'gi');
            return element.match(resizeClass);
          });
      }
      return r;
    },
    isResizeElement(model = null, findPatternColumn = null) {
      let r = false;
      if (
        model &&
        model.attributes &&
        model.attributes.classes &&
        model.attributes.classes.models &&
        model.attributes.classes.models.length > 0 &&
        findPatternColumn
      ) {
        const elClass = this.getClasses();
        r = elClass.find(element => {
          const resizeClass = new RegExp(`${findPatternColumn}[\\d]`, 'gi');
          return element.match(resizeClass);
        });
      }
      return r;
    },
    initToolbar() {
      const { em } = this;
      const model = this;
      const ppfx = (em && em.getConfig('stylePrefix')) || '';
      const buttons = em && em.getConfig('domComponents');
      const { toolbar } = buttons;
      const upperable = toolbar.upperable || false;
      const draggable = toolbar.draggable || false;
      const removable = toolbar.removable || false;
      const copyable = toolbar.copyable || false;
      const expandable = toolbar.expandable || false;
      const reducible = toolbar.reducible || false;
      const parentClassResize = toolbar.parentClassResize || false;
      const findPatternColumn = toolbar.findPatternColumn || false;

      if (!model.get('toolbar')) {
        const tb = [];
        if (model.collection && upperable) {
          tb.push({
            attributes: { class: 'fa fa-arrow-up' },
            command: 'select-parent'
          });
        }
        if (model.get('draggable') && draggable) {
          tb.push({
            attributes: {
              class: `fa fa-arrows ${ppfx}no-touch-actions`,
              draggable: true
            },
            command: 'tlb-move'
          });
        }
        if (model.get('copyable') && copyable) {
          tb.push({
            attributes: { class: 'fa fa-clone' },
            command: 'tlb-clone'
          });
        }
        if (model.get('removable') && removable) {
          tb.push({
            attributes: { class: 'fa fa-trash' },
            command: 'tlb-delete'
          });
        }
        if (
          model.attributes.tagName === 'div' &&
          expandable &&
          parentClassResize &&
          findPatternColumn
        ) {
          const isResizeElement = this.isResizeElement(
            model,
            findPatternColumn
          );
          const validResizeParent = this.validResizeParent(parentClassResize);
          if (isResizeElement && validResizeParent) {
            tb.push({
              attributes: { class: 'fa fa-plus' },
              command: editor => {
                const collection = this.collection;
                if (collection && collection.parent) {
                  const allChldrens = collection.parent.get('components');
                  if (
                    allChldrens &&
                    allChldrens.models &&
                    allChldrens.models.length > 0
                  ) {
                    const siblings = allChldrens.models;
                    siblings.splice(allChldrens.models.indexOf(this), 1);
                    console.log('-> ', siblings);
                  }
                }
                // falta hacer el resize de sus hermanos se puede hacer con la function hecha en caracas (va en el command)
                // aca tengo que colocar el resize de la columnas
                window.emodel = model;
                window.eeditor = editor;
                window.ethis = this;
                // editor.getSelected().parent().attributes.classes.models;
              }
            });
          }
        }
        if (
          model.attributes.tagName === 'div' &&
          reducible &&
          parentClassResize &&
          findPatternColumn
        ) {
          const isResizeElement = this.isResizeElement(
            model,
            findPatternColumn
          );
          const validResizeParent = this.validResizeParent(parentClassResize);
          if (isResizeElement && validResizeParent) {
            tb.push({
              attributes: { class: 'fa fa-minus' },
              command: (editor, nose, options) => {
                // aca tengo que colocar el resize de la columnas
                console.log(model.attributes);
                console.log('reducible', editor, nose, options);
                window.test = model;
                window.editor = editor;
                editor.getSelected().parent().attributes.classes.models;
              }
            });
          }
        }
        model.set('toolbar', tb);
      }
    },
    loadTraits(traits, opts = {}) {
      traits = traits || this.get('traits');
      if (!(traits instanceof Traits)) {
        const trt = new Traits([], this.opt);
        trt.setTarget(this);
        if (traits.length) {
          traits.forEach(tr => tr.attributes && delete tr.attributes.value);
          trt.add(traits);
        }
        this.set('traits', trt, opts);
      }
      return this;
    },
    getTrait(id) {
      return this.get('traits').filter(
        trait => trait.get('id') === id || trait.get('name') === id
      )[0];
    },
    normalizeClasses(arr) {
      const res = [];
      const em = this.em;
      if (!em) return;
      const clm = em.get('SelectorManager');
      if (!clm) {
        return;
      }
      arr.forEach(val => {
        let name = '';
        if (typeof val === 'string') name = val;
        else name = val.name;
        const model = clm.add(name);
        res.push(model);
      });
      return res;
    },
    clone() {
      const em = this.em;
      const style = this.getStyle();
      const attr = { ...this.attributes };
      const opts = { ...this.opt };
      attr.attributes = { ...attr.attributes };
      delete attr.attributes.id;
      attr.components = [];
      attr.classes = [];
      attr.traits = [];
      this.get('components').each((md, i) => {
        attr.components[i] = md.clone();
      });
      this.get('traits').each((md, i) => {
        attr.traits[i] = md.clone();
      });
      this.get('classes').each((md, i) => {
        attr.classes[i] = md.get('name');
      });
      attr.status = '';
      attr.view = '';
      opts.collection = null;
      if (em && em.getConfig('avoidInlineStyle') && !isEmpty(style)) {
        attr.style = style;
      }

      const cloned = new this.constructor(attr, opts);
      const event = 'component:clone';
      if (em && em.trigger) {
        em.trigger(event, cloned);
      }
      this.trigger(event, cloned);
      return cloned;
    },
    getName() {
      const customName = this.get('name') || this.get('custom-name');
      let tag = this.get('tagName');
      tag = tag == 'div' ? 'box' : tag;
      let name = this.get('type') || tag;
      name = name.charAt(0).toUpperCase() + name.slice(1);
      return customName || name;
    },
    getIcon() {
      const icon = this.get('icon');
      return icon ? `${icon} ` : '';
    },
    toHTML(opts = {}) {
      const model = this;
      const attrs = [];
      const classes = [];
      const tag = model.get('tagName');
      const sTag = model.get('void');
      const customAttr = opts.attributes;
      let attributes = this.getAttrToHTML();

      // Get custom attributes if requested
      if (customAttr) {
        if (isFunction(customAttr)) {
          attributes = customAttr(model, attributes) || {};
        } else if (isObject(customAttr)) {
          attributes = customAttr;
        }
      }
      for (const attr in attributes) {
        const val = attributes[attr];
        const value = isString(val) ? val.replace(/"/g, '&quot;') : val;
        if (!isUndefined(value)) {
          if (isBoolean(value)) {
            value && attrs.push(attr);
          } else {
            attrs.push(`${attr}="${value}"`);
          }
        }
      }

      const attrString = attrs.length ? ` ${attrs.join(' ')}` : '';
      let code = `<${tag}${attrString}${sTag ? '/' : ''}>${model.get(
        'content'
      )}`;
      model.get('components').each(comp => (code += comp.toHTML(opts)));
      if (!sTag) {
        code += `</${tag}>`;
      }
      return code;
    },
    getAttrToHTML() {
      const attr = this.getAttributes();
      delete attr.style;
      return attr;
    },
    toJSON(...args) {
      const obj = Backbone.Model.prototype.toJSON.apply(this, args);
      obj.attributes = this.getAttributes();
      delete obj.attributes.class;
      delete obj.toolbar;
      if (this.em.getConfig('avoidDefaults')) {
        const defaults = result(this, 'defaults');
        forEach(defaults, (value, key) => {
          if (['type', 'content'].indexOf(key) === -1 && obj[key] === value) {
            delete obj[key];
          }
        });
        if (isEmpty(obj.type)) {
          delete obj.type;
        }
        forEach(['attributes', 'style'], prop => {
          if (isEmpty(defaults[prop]) && isEmpty(obj[prop])) {
            delete obj[prop];
          }
        });
        forEach(['classes', 'components'], prop => {
          if (isEmpty(defaults[prop]) && !obj[prop].length) {
            delete obj[prop];
          }
        });
      }
      return obj;
    },
    getId() {
      const attrs = this.get('attributes') || {};
      return attrs.id || this.ccid || this.cid;
    },
    setId(id) {
      const attrs = { ...this.get('attributes') };
      attrs.id = id;
      this.set('attributes', attrs);
      return this;
    },
    getEl() {
      return this.view && this.view.el;
    },
    getView() {
      return this.view;
    },
    getScriptString(script) {
      let scr = script || this.get('script');
      if (!scr) {
        return scr;
      }
      if (typeof scr === 'function') {
        let scrStr = scr.toString().trim();
        scrStr = scrStr
          .replace(/^function[\s\w]*\(\)\s?\{/, '')
          .replace(/\}$/, '');
        scr = scrStr.trim();
      }
      const config = this.em.getConfig();
      const tagVarStart = escapeRegExp(config.tagVarStart || '{[ ');
      const tagVarEnd = escapeRegExp(config.tagVarEnd || ' ]}');
      const reg = new RegExp(`${tagVarStart}([\\w\\d-]*)${tagVarEnd}`, 'g');
      scr = scr.replace(reg, (match, v) => {
        this.scriptUpdated();
        return this.attributes[v] || '';
      });
      return scr;
    },
    emitUpdate(property, ...args) {
      const em = this.em;
      const event = `component:update${property ? `:${property}` : ''}`;
      if (property) {
        this.updated(
          property,
          property && this.get(property),
          property && this.previous(property),
          ...args
        );
      }
      this.trigger(event, ...args);
      if (em && em.trigger) {
        em.trigger(event, this, ...args);
      }
    },
    onAll(clb) {
      if (isFunction(clb)) {
        clb(this);
        this.components().forEach(model => model.onAll(clb));
      }
      return this;
    },
    remove() {
      return this.collection.remove(this);
    },
    resetId(opts = {}) {
      const { em } = this;
      const oldId = this.getId();
      if (!oldId) return;
      const newId = Component.createId(this);
      this.setId(newId);
      const rule = em && em.get('CssComposer').getIdRule(oldId);
      const selector = rule && rule.get('selectors').at(0);
      if (selector && selector.set) {
        selector.set('name', newId);
      }
      return this;
    }
  },
  {
    isComponent(el) {
      return { tagName: el.tagName ? el.tagName.toLowerCase() : '' };
    },
    createId(model) {
      componentIndex++;
      const ilen = componentIndex.toString().length + 2;
      const uid = (Math.random() + 1.1).toString(36).slice(-ilen);
      const nextId = `i${uid}`;
      componentList[nextId] = model;
      return nextId;
    },
    getList() {
      return componentList;
    }
  }
);

module.exports = Component;
