import { isUndefined, isArray, contains, toArray, keys } from 'underscore';
import { getModel } from 'utils/mixins';

const deps = [
  require('utils'),
  require('keymaps'),
  require('undo_manager'),
  require('storage_manager'),
  require('device_manager'),
  require('parser'),
  require('style_manager'),
  require('selector_manager'),
  require('modal_dialog'),
  require('code_manager'),
  require('panels'),
  require('rich_text_editor'),
  require('asset_manager'),
  require('css_composer'),
  require('trait_manager'),
  require('dom_components'),
  require('navigator'),
  require('canvas'),
  require('commands'),
  require('block_manager')
];

const Backbone = require('backbone');

const { Collection } = Backbone;
let timedInterval;

require('utils/extender')({
  Backbone,
  $: Backbone.$
});

const $ = Backbone.$;
const logs = {
  debug: console.log,
  info: console.info,
  warning: console.warn,
  error: console.error
};
module.exports = Backbone.Model.extend({
  defaults() {
    return {
      editing: 0,
      selected: new Collection(),
      clipboard: null,
      designerMode: false,
      componentHovered: null,
      previousModel: null,
      changesCount: 0,
      storables: [],
      modules: [],
      toLoad: [],
      opened: {},
      device: ''
    };
  },
  initialize(c = {}) {
    this.config = c;
    this.set('Config', c);
    this.set('modules', []);
    this.set('toLoad', []);
    this.set('storables', []);
    const el = c.el;
    const log = c.log;
    const toLog = log === true ? keys(logs) : isArray(log) ? log : [];
    if (el && c.fromElement) this.config.components = el.innerHTML;
    this.attrsOrig = el
      ? toArray(el.attributes).reduce((res, next) => {
          res[next.nodeName] = next.nodeValue;
          return res;
        }, {})
      : '';
    deps.forEach(name => this.loadModule(name));
    this.on('change:componentHovered', this.componentHovered, this);
    this.on('change:changesCount', this.updateChanges, this);
    toLog.forEach(e => this.listenLog(e));
    [{ from: 'change:selectedComponent', to: 'component:toggled' }].forEach(
      event => {
        const eventFrom = event.from;
        const eventTo = event.to;
        this.listenTo(this, eventFrom, (...args) => {
          this.trigger(eventTo, ...args);
          this.logWarning(
            `The event '${eventFrom}' is deprecated, replace it with '${eventTo}'`
          );
        });
      }
    );
  },
  listenLog(event) {
    this.listenTo(this, `log:${event}`, logs[event]);
  },
  getConfig(prop) {
    const config = this.config;
    return isUndefined(prop) ? config : config[prop];
  },
  loadOnStart(clb = null) {
    const sm = this.get('StorageManager');
    this.get('toLoad').forEach(module => {
      module.onLoad();
    });
    const postLoad = () => {
      const modules = this.get('modules');
      modules.forEach(module => module.postLoad && module.postLoad(this));
      if (clb) {
        clb();
      }
    };

    if (sm && sm.canAutoload()) {
      this.load(postLoad);
    } else {
      postLoad();
    }
  },
  updateChanges() {
    const stm = this.get('StorageManager');
    const changes = this.get('changesCount');
    if (this.config.noticeOnUnload) {
      window.onbeforeunload = changes ? e => 1 : null;
    }
    if (stm.isAutosave() && changes >= stm.getStepsBeforeSave()) {
      this.store();
    }
  },
  loadModule(moduleName) {
    const c = this.config;
    const Mod = new moduleName();
    const name = Mod.name.charAt(0).toLowerCase() + Mod.name.slice(1);
    const cfg = c[name] || c[Mod.name] || {};
    cfg.pStylePrefix = c.pStylePrefix || '';
    const sm = this.get('StorageManager');
    if (Mod.storageKey && Mod.store && Mod.load && sm) {
      cfg.stm = sm;
      const storables = this.get('storables');
      storables.push(Mod);
      this.set('storables', storables);
    }
    cfg.em = this;
    Mod.init({ ...cfg });
    !Mod.private && this.set(Mod.name, Mod);
    Mod.onLoad && this.get('toLoad').push(Mod);
    this.get('modules').push(Mod);
    return this;
  },
  init(editor) {
    this.set('Editor', editor);
  },
  getEditor() {
    return this.get('Editor');
  },
  handleUpdates(model, val, opt = {}) {
    if (opt.temporary) {
      return;
    }
    timedInterval && clearInterval(timedInterval);
    timedInterval = setTimeout(() => {
      if (!opt.avoidStore) {
        this.set('changesCount', this.get('changesCount') + 1, opt);
      }
    }, 0);
  },
  componentHovered(editor, component, options) {
    const prev = this.previous('componentHovered');
    prev && this.trigger('component:unhovered', prev, options);
    component && this.trigger('component:hovered', component, options);
  },
  getSelected() {
    return this.get('selected').last();
  },
  getSelectedAll() {
    return this.get('selected').models;
  },
  setSelected(el, opts = {}) {
    const multiple = isArray(el);
    const els = multiple ? el : [el];
    const selected = this.get('selected');
    multiple && this.removeSelected(selected.filter(s => !contains(els, s)));
    els.forEach(el => {
      const model = getModel(el, $);
      if (model && !model.get('selectable')) return;
      !multiple && this.removeSelected(selected.filter(s => s !== model));
      this.addSelected(model, opts);
    });
  },
  addSelected(el, opts = {}) {
    const model = getModel(el, $);
    const models = isArray(model) ? model : [model];
    models.forEach(model => {
      if (model && !model.get('selectable')) return;
      const selected = this.get('selected');
      opts.forceChange && selected.remove(model, opts);
      selected.push(model, opts);
    });
  },
  removeSelected(el, opts = {}) {
    this.get('selected').remove(getModel(el, $), opts);
  },
  toggleSelected(el, opts = {}) {
    const model = getModel(el, $);
    const models = isArray(model) ? model : [model];
    models.forEach(model => {
      if (this.get('selected').contains(model)) {
        this.removeSelected(model, opts);
      } else {
        this.addSelected(model, opts);
      }
    });
  },
  setHovered(el, opts = {}) {
    const model = getModel(el, $);
    if (model && !model.get('hoverable')) return;
    opts.forceChange && this.set('componentHovered', '');
    this.set('componentHovered', model, opts);
  },
  setComponents(components) {
    return this.get('DomComponents').setComponents(components);
  },
  getComponents() {
    const cmp = this.get('DomComponents');
    const cm = this.get('CodeManager');
    if (!cmp || !cm) return;
    const wrp = cmp.getComponents();
    return cm.getCode(wrp, 'json');
  },
  setStyle(style) {
    const rules = this.get('CssComposer').getAll();
    for (let i = 0, len = rules.length; i < len; i++) rules.pop();
    rules.add(style);
    return this;
  },
  getStyle() {
    return this.get('CssComposer').getAll();
  },
  getHtml() {
    const config = this.config;
    const exportWrapper = config.exportWrapper;
    const wrappesIsBody = config.wrappesIsBody;
    const js = config.jsInHtml ? this.getJs() : '';
    const wrp = this.get('DomComponents').getComponent();
    let html = this.get('CodeManager').getCode(wrp, 'html', {
      exportWrapper,
      wrappesIsBody
    });
    html += js ? `<script>${js}</script>` : '';
    return html;
  },
  getCss(opts = {}) {
    const config = this.config;
    const wrappesIsBody = config.wrappesIsBody;
    const avoidProt = opts.avoidProtected;
    const keepUnusedStyles = !isUndefined(opts.keepUnusedStyles)
      ? opts.keepUnusedStyles
      : config.keepUnusedStyles;
    const cssc = this.get('CssComposer');
    const wrp = this.get('DomComponents').getComponent();
    const protCss = !avoidProt ? config.protectedCss : '';
    return (
      protCss +
      this.get('CodeManager').getCode(wrp, 'css', {
        cssc,
        wrappesIsBody,
        keepUnusedStyles
      })
    );
  },
  getJs() {
    const wrp = this.get('DomComponents').getWrapper();
    return this.get('CodeManager')
      .getCode(wrp, 'js')
      .trim();
  },
  store(clb) {
    const sm = this.get('StorageManager');
    const store = {};
    if (!sm) return;
    this.get('storables').forEach(m => {
      const obj = m.store(1);
      for (const el in obj) store[el] = obj[el];
    });
    sm.store(store, res => {
      clb && clb(res);
      this.set('changesCount', 0);
      this.trigger('storage:store', store);
    });
    return store;
  },
  load(clb = null) {
    this.getCacheLoad(1, res => {
      this.get('storables').forEach(module => module.load(res));
      clb && clb(res);
    });
  },
  getCacheLoad(force, clb) {
    const f = force ? 1 : 0;
    if (this.cacheLoad && !f) return this.cacheLoad;
    const sm = this.get('StorageManager');
    const load = [];
    if (!sm) return {};
    this.get('storables').forEach(m => {
      let key = m.storageKey;
      key = typeof key === 'function' ? key() : key;
      const keys = key instanceof Array ? key : [key];
      keys.forEach(k => {
        load.push(k);
      });
    });
    sm.load(load, res => {
      this.cacheLoad = res;
      clb && clb(res);
      setTimeout(() => this.trigger('storage:load', res), 0);
    });
  },
  getDeviceModel() {
    const name = this.get('device');
    return this.get('DeviceManager').get(name);
  },
  runDefault(opts = {}) {
    const command = this.get('Commands').get(this.config.defaultCommand);
    if (!command || this.defaultRunning) return;
    command.stop(this, this, opts);
    command.run(this, this, opts);
    this.defaultRunning = 1;
  },
  stopDefault(opts = {}) {
    const command = this.get('Commands').get(this.config.defaultCommand);
    if (!command) return;
    command.stop(this, this, opts);
    this.defaultRunning = 0;
  },
  refreshCanvas() {
    this.set('canvasOffset', this.get('Canvas').getOffset());
  },
  clearSelection(win) {
    const w = win || window;
    w.getSelection().removeAllRanges();
  },
  getCurrentMedia() {
    const config = this.config;
    const device = this.getDeviceModel();
    const condition = config.mediaCondition;
    const preview = config.devicePreviewMode;
    const width = device && device.get('widthMedia');
    return device && width && !preview ? `(${condition}: ${width})` : '';
  },
  getWrapper() {
    return this.get('DomComponents').getWrapper();
  },
  getDirtyCount() {
    return this.get('changesCount');
  },
  destroyAll() {
    const {
      DomComponents,
      CssComposer,
      UndoManager,
      Panels,
      Canvas,
      Keymaps
    } = this.attributes;
    DomComponents.clear();
    CssComposer.clear();
    UndoManager.clear().removeAll();
    Panels.getPanels().reset();
    Canvas.getCanvasView().remove();
    Keymaps.removeAll();
    this.view.remove();
    this.stopListening();
    $(this.config.el)
      .empty()
      .attr(this.attrsOrig);
  },
  setEditing(value) {
    this.set('editing', value);
    return this;
  },
  isEditing() {
    return !!this.get('editing');
  },
  log(msg, opts = {}) {
    const { ns, level = 'debug' } = opts;
    this.trigger('log', msg, opts);
    level && this.trigger(`log:${level}`, msg, opts);
    if (ns) {
      const logNs = `log-${ns}`;
      this.trigger(logNs, msg, opts);
      level && this.trigger(`${logNs}:${level}`, msg, opts);
    }
  },
  logInfo(msg, opts) {
    this.log(msg, { ...opts, level: 'info' });
  },
  logWarning(msg, opts) {
    this.log(msg, { ...opts, level: 'warning' });
  },
  logError(msg, opts) {
    this.log(msg, { ...opts, level: 'error' });
  },
  data(el, name, value) {
    const varName = '_rae-data';

    if (!el[varName]) {
      el[varName] = {};
    }

    if (isUndefined(value)) {
      return el[varName][name];
    }
    el[varName][name] = value;
  }
});
